"""Pipe debuging utilities"""

from pipe import Pipe, batched, islice

__all__ = ["debug_eager"]


("Hello world!"
 | as_list()
 | take_while(lambda x: x != " ")
 | debug_eager("Characters:")
 | map(ord)
 | debug_eager("Codepoints:")
 | as_list())


@Pipe
def debug_eager(iterable, label: str | None = None, max_count: int = 100):
    """
    Debugging helper. Prints label if given, then eagerly fetches items from iterable,
    prints them, then passes them on. This is similar to tee(), but tee() is
    completely lazy, and will print items one-by-one, whereas this function will 

    For safety, only up to max_count items will be printed, to avoid infinite loops.

    Note that although this function is eager once it starts fetching the items, it's
    still a generator and thus won't actually do anything until something starts to
    iterate the pipeline (e.g. as_list()).

    Useful for figuring out why the pipeline doesn't produce all the expected elements,
    for example with take_while().

    >>> ("Hello world!"
    ...  | as_list()
    ...  | take_while(lambda x: x != " ")
    ...  | debug_eager("Characters:")
    ...  | map(ord)
    ...  | debug_eager("Codepoints:")
    ...  | as_list())
    Characters:
    H
    e
    l
    l
    o
    Codepoints:
    72
    101
    108
    108
    111
    [72, 101, 108, 108, 111]
    """
    iterable = iter(iterable)
    items = list(iterable | islice(None, max_count))
    if label is not None:
        print(label)
    for item in items:
        print(item)
    yield from items
    yield from iterable



"""
Simple wrappers around existing functions
"""

from functools import reduce

from pipe import Pipe

__all__ = [
    "as_list",
    "as_dict",
    "as_tuple",
    "as_sum",
    "join",
    "reduce",
]


def as_list() -> Pipe:
    """Convert pipe's iterable to list. Simple shorthand for Pipe(list)"""
    return Pipe(list)


def as_dict() -> Pipe:
    """Convert pipe's iterable to dict. Simple shorthand for Pipe(dict)"""
    return Pipe(dict)


def as_tuple() -> Pipe:
    """Convert pipe's iterable to tuple. Simple shorthand for Pipe(tuple)"""
    return Pipe(tuple)


@Pipe
def as_sum(iterable, /, start=0):
    """
    Return sum of the pipe's iterable. Wraps builtin sum() as a Pipe, same as add()
    from Pipe 1.x
    """
    return sum(iterable, start=start)


def join(sep: str):
    """
    Join the iterable into a string with sep. Wraps sep.join as a Pipe
    """
    return Pipe(sep.join)


@Pipe
def reduce(iterable, function, *args, **kwargs):
    """
    Reduce the pipe's iterable with function. Wraps functools.reduce() as a Pipe
    """
    return reduce(function, iterable, *args, **kwargs)
